/**
  *  \file client/dialogs/keymapdlg.cc
  *  \brief Keymap Debugger Dialog
  *
  *  (c) 2001-2011 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/dialogs/keymapdlg.h"
#include "ui/widgets/multilinestatic.h"
#include "ui/layout.h"
#include "ui/window.h"
#include "util/nls.h"
#include "cpluslib/format.h"
#include "util/atom.h"
#include "ui/uiglobal.h"
#include "gfx/palette.h"
#include "ui/widgets/button.h"

namespace {
    class WKeymapDebugger : public UIWindow {
        ObjectHolder h;
        IntKeymapRef keymap;
        UIMultilineStatic* keymaptree;
        UIMultilineStatic* response;
     public:
        WKeymapDebugger(IntKeymapRef keymap);
        void init();
        bool handleEvent(const UIEvent& event, bool second_pass);
    };
}

/** Format keymap tree into a string. Given a keymap, list all its parents in a multiline string,
    starting with the keymap at the left, and all parents indented.
    \param keymap The keymap to describe
    \param nlines Number of lines to generate */
static string_t
formatKeymapTree(IntKeymapRef keymap, int nlines)
{
    // Start with a single line listing the current keymap
    string_t result = keymap->getName();
    result += '\n';
    --nlines;

    // This implements kind-of a stack for pseudo recursion. Doing it
    // iteratively, though, allows us to easily keep track of the number
    // of lines we generated.
    enum { DEPTH = 5 };
    IntKeymapRef parents[DEPTH];
    uint32_t indexes[DEPTH];
    int level = 0;

    parents[0] = keymap;
    indexes[0] = 0;
    while (nlines > 0 && level >= 0) {
        if (indexes[level] >= parents[level]->getNumDirectParents()) {
            // Finished with this level
            --level;
        } else if (level >= DEPTH-1) {
            // Maximum level, so display just a placeholder
            result.append(2*level+2, ' ');
            result += "...\n";
            --nlines;
            --level;
        } else {
            // Fetch one parent, and add it
            IntKeymapRef p = parents[level]->getDirectParent(indexes[level]++);
            result.append(2*level+2, ' ');
            result += p->getName();
            result += '\n';
            ++level;
            indexes[level] = 0;
            parents[level] = p;
            --nlines;
        }
    }

    return result;
}

WKeymapDebugger::WKeymapDebugger(IntKeymapRef keymap)
    : UIWindow(_("Keymap Debugger"), BLUE_WINDOW, COLOR_GRAY),
      keymap(keymap),
      keymaptree(0),
      response(0)
{
    init();
}

void
WKeymapDebugger::init()
{
    // VBox
    //   HBox
    //     keymaptree
    //     response
    //   HBox
    //     UISpacer
    //     "Close"

    setLayout(UIVBoxLayout::instance5);

    UIGridLayout& g1layout = h.add(new UIGridLayout(2));
    UIGroup& g1 = h.add(new UIGroup(g1layout));
    g1layout.setForcedCellSize(220, 12*font_heights[FONT_NORMAL]);

    keymaptree = &h.add(new UIMultilineStatic(FONT_NORMAL, formatKeymapTree(keymap, 12), 220));
    g1.add(*keymaptree);

    response = &h.add(new UIMultilineStatic(FONT_NORMAL, _("Press the key for which you want information."), 220));
    g1.add(*response);
    
    UIGroup& g2 = h.add(new UIGroup(UIHBoxLayout::instance5));
    g2.add(h.add(new UISpacer()));
    g2.add(h.add(new UIButton(0, _("Close"), 0)).addStop());

    add(g1);
    add(g2);
    pack();
}

bool
WKeymapDebugger::handleEvent(const UIEvent& event, bool second_pass)
{
    if (event.type == ev_Key && !second_pass && getClassOfKey(event.key.code & ~ss_AllMod) != KeyModifier) {
        // Shift-Escape always stops
        if (event.key.code == ss_Shift + SDLK_ESCAPE) {
            stop(0);
            return true;
        }

        // Attempt to look up other key
        IntKeymapRef where = 0;
        uint32_t cmd = keymap->lookupCommand(event.key.code, where);
        string_t message = format(_("Key %s:\n"), formatKey(event.key.code));
        bool used;
        if (cmd != 0 && where != 0) {
            // Bound
            message += format(_("Bound in keymap %s.\n"), where->getName());
            if (isAtom(cmd)) {
                message += format(_("Command:\n  %s\n"), atomStr(cmd));
            } else {
                message += format(_("Generates internal command %d.\n"), cmd);
            }
            used = true;
        } else {
            // Not bound
            if (where) {
                // Key is bound to 0 in a keymap
                message += format(_("This key is unbound by %s.\n"), where->getName());
                used = true;
            } else {
                // Not bound at all
                message += _("This key is not bound.\n");
                if ((event.key.code & ~ss_AllMod) == vk_Unicode)
                    message += _("This key produces a character not available in your game character set, "
                                 "which cannot be bound.\n");
                used = false;
            }
        }

        // Process result
        if (event.key.code == SDLK_ESCAPE) {
            if (used)
                message += _("Press Shift-ESC to close this window.\n");
            else
                stop(0);
        }
        response->setText(message);
        return true;
    } else {
        return UIWindow::handleEvent(event, second_pass);
    }
}

/** Do keymap debugger dialog.
    \param keymap Keymap to analyze */
void
doKeymapDebugger(IntKeymapRef keymap)
{
    // If the keymap does not exist, fail silently
    if (!keymap)
        return;

    WKeymapDebugger w(keymap);
    ui_root->centerWidget(w);
    ui_root->runModal(w);
}
