/**
  *  \file client/tiles/planetmining.cc
  *
  *  (c) 2001-2011 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/tiles/planetmining.h"
#include "cpluslib/format.h"
#include "game/planet.h"
#include "game/tables.h"
#include "gfx/gfx.h"
#include "gfx/palette.h"
#include "ui/uiglobal.h"
#include "ui/util.h"
#include "util/misc.h"
#include "util/nls.h"
#include "ui/table.h"

WPlanetMiningTile::WPlanetMiningTile(GPlanet& planet, GCargoType ele)
    : WTile(GfxPoint(295, 4 * font_heights[FONT_NORMAL] + font_heights[FONT_TITLE] + 5)),
      planet(planet), ele(ele),
      mines(planet.getStructures(str_Mines))
{ }

static void
drawPercentBar(GfxCanvas& can, GfxRect rect, long fill, int divi, int color)
{
    int split = rect.w * fill / (100*divi);
    if (fill && !split)
        split = 1;
    if (split > rect.w)
        split = rect.w;

    drawSolidBar(can, GfxRect(rect.x, rect.y, split, rect.h), standard_colors[color]);
    drawSolidBar(can, GfxRect(rect.x+split, rect.y, rect.w-split, rect.h), standard_colors[COLOR_SHIELD+3]);
}

void
WPlanetMiningTile::drawData(GfxCanvas& can)
{
    const int wi = extent.w/2 - 5;
    int y = 3;

    GfxRect in = getInnerSize(0, 0);

    drawFrameDown(can, in);

    GfxContext ctx;
    ctx.useCanvas(can);
    ctx.useColorScheme(getSkin());
    ctx.useFont(*fonts[FONT_NORMAL]);
    ctx.setColor(UIColor::tc_Heading);
    ctx.setSolidBackground();

    outTextF(ctx, 5, y, wi, getCargoName(ele));
    y += font_heights[FONT_NORMAL] + 3;

    ctx.setColor(UIColor::tc_Static);

    if (planet.getOreGround(ele).isKnown()
        && planet.getOreDensity(ele).isKnown()
        && planet.isCargoAmountKnown(ele))
    {
        /* Mining */
        if (mines.isKnown()) {
            int amount   = planet.getMiningCapacity(ele, mines);
            int32 ground = planet.getOreGround(ele);

            // Show mining capacity
            if (amount > ground)
                ctx.setColor(UIColor::tc_Yellow);
            else
                ctx.setColor(UIColor::tc_Green);
            ctx.setTextAlign(2, 0);
            outTextF(ctx, in.w-5, 3, wi, format(_("%d kt/turn"), amount));

            // Show mining time
            string_t timestr;
            if (amount != 0) {
                int32 time = (ground + (amount-1)) / amount;
                if (time > 30)
                    timestr = _(">30 turns");
                else
                    timestr = format(_("%d turn%!1{s%}"), time);
            }
            outTextF(ctx, in.w-5, 3+font_heights[FONT_NORMAL], wi, timestr);
            ctx.setTextAlign(0, 0);
        }

        /* Ore */
        ctx.setColor(UIColor::tc_Static);
        if (planet.getOreGround(ele))
            outTextF(ctx, 5, y, wi,
                     format("%s, %s", getMineralMassClassName(planet.getOreGround(ele)),
                            getMineralDensityClassName(planet.getOreDensity(ele))));
        else
            outTextF(ctx, 5, y, wi, _("none"));
        y += font_heights[FONT_NORMAL];

        outText(ctx, 5, y, _("mined"));
        drawPercentBar(can, GfxRect(in.x+60, in.y+y, 150, 15),
                       planet.getCargo(ele), 82, COLOR_WHITE);
        ctx.setTextAlign(2, 0);
        outTextF(ctx, in.w - 25, y, 60, numToString(planet.getCargo(ele)));
        ctx.setTextAlign(0, 0);
        outTextF(ctx, in.w - 25, y, 20, " kt");
        y += font_heights[FONT_NORMAL];

        outText(ctx, 5, y, _("ground"));
        drawPercentBar(can, GfxRect(in.x+60, in.y+y, 150, 15),
                       planet.getOreGround(ele), 82, COLOR_WHITE);
        ctx.setTextAlign(2, 0);
        outTextF(ctx, in.w - 25, y, 60, numToString(planet.getOreGround(ele)));
        ctx.setTextAlign(0, 0);
        outTextF(ctx, in.w - 25, y, 20, " kt");
        y += font_heights[FONT_NORMAL];

        outText(ctx, 5, y, _("density"));
        drawPercentBar(can, GfxRect(in.x+60, in.y+y, 150, 15),
                       planet.getOreDensity(ele), 1, COLOR_GRAY);
        ctx.setTextAlign(2, 0);
        outTextF(ctx, in.w - 25, y, 60, numToString(planet.getOreDensity(ele)));
        ctx.setTextAlign(0, 0);
        outTextF(ctx, in.w - 25, y, 20, " %");
        y += font_heights[FONT_NORMAL];
    } else if (planet.isCargoAmountKnown(ele)) {
        if (planet.getCargo(ele))
            outTextF(ctx, 5, y, in.w-10, getMineralMassClassName(planet.getCargo(ele)));
        else
            outTextF(ctx, 5, y, in.w-10, _("none"));
        y += font_heights[FONT_NORMAL];

        outText(ctx, 5, y, _("total"));
        drawPercentBar(can, GfxRect(in.x+60, in.y+y, 150, 15),
                       planet.getCargo(ele), 82, COLOR_WHITE);
        ctx.setTextAlign(2, 0);
        outTextF(ctx, in.w - 25, y, 60, numToString(planet.getCargo(ele)));
        ctx.setTextAlign(0, 0);
        outTextF(ctx, in.w - 25, y, 20, " kt");
        y += font_heights[FONT_NORMAL];

        outTextFormatted(ctx, in.x + 5, in.y + y, in.w - 10,
                         _("(no information on density and mined/ground masses available.)"));
    }
}

void
WPlanetMiningTile::setNumMines(mp16_t mines)
{
    this->mines = mines;
    drawWidget();
}




WNarrowPlanetMineralTile::WNarrowPlanetMineralTile(GObjectType& type)
    : WObjectTypeObserverTile(GfxPoint(250, 5 * font_heights[FONT_NORMAL]), type)
{ }

void
WNarrowPlanetMineralTile::drawData(GfxCanvas& can)
{
    GfxContext ctx;
    ctx.useCanvas(can).useColorScheme(getSkin());

    GPlanet* pl = dynamic_cast<GPlanet*>(getObject());
    if (!pl) {
        drawBackground(ctx, getInnerSize(0, 0));
        return;
    }

    enum { N = 4 };
    static const GCargoType eles[N] = { el_Neutronium, el_Tritanium, el_Duranium, el_Molybdenum };

    UITableWriter tw(ctx, getInnerSize(0, 0), FONT_NORMAL);
    tw.newLColumn();
    tw.writeFixedCell(UIColor::tc_Static, string_t());
    tw.writeFixedCell(UIColor::tc_Static, _("Neu:"));
    tw.writeFixedCell(UIColor::tc_Static, _("Tri:"));
    tw.writeFixedCell(UIColor::tc_Static, _("Dur:"));
    tw.writeFixedCell(UIColor::tc_Static, _("Mol:"));

    /* Mined */
    tw.newRColumn(fonts[FONT_NORMAL]->getTextWidth("9,999,999"));
    tw.writeFixedCell(UIColor::tc_Static, _("Mined"));
    for (int i = 0; i < N; ++i) {
        if (pl->isCargoAmountKnown(eles[i])) {
            tw.writeVariableCell(UIColor::tc_Green, numToString(pl->getCargoRaw(eles[i])));
        } else {
            tw.writeVariableCell(UIColor::tc_Green, string_t());
        }
    }
    tw.newLColumn(fonts[FONT_NORMAL]->getTextWidth("kt"));
    tw.writeFixedCell(UIColor::tc_Static, string_t());
    for (int i = 0; i < N; ++i) {
        if (pl->isCargoAmountKnown(eles[i])) {
            tw.writeVariableCell(UIColor::tc_Green, "kt");
        } else {
            tw.writeVariableCell(UIColor::tc_Green, string_t());
        }
    }

    /* Ground */
    tw.newRColumn(fonts[FONT_NORMAL]->getTextWidth("9,999,999"));
    tw.writeFixedCell(UIColor::tc_Static, _("Ground"));
    for (int i = 0; i < N; ++i) {
        if (pl->getOreGround(eles[i]).isKnown()) {
            tw.writeVariableCell(UIColor::tc_Green, numToString(pl->getOreGround(eles[i])));
        } else {
            tw.writeVariableCell(UIColor::tc_Green, string_t());
        }
    }
    tw.newLColumn(fonts[FONT_NORMAL]->getTextWidth("kt"));
    tw.writeFixedCell(UIColor::tc_Static, string_t());
    for (int i = 0; i < N; ++i) {
        if (pl->getOreGround(eles[i]).isKnown()) {
            tw.writeVariableCell(UIColor::tc_Green, "kt");
        } else {
            tw.writeVariableCell(UIColor::tc_Green, string_t());
        }
    }

    /* Density */
    tw.newRColumn(fonts[FONT_NORMAL]->getTextWidth("100%"));
    tw.writeFixedCell(UIColor::tc_Static, string_t());
    for (int i = 0; i < N; ++i) {
        if (pl->getOreDensity(eles[i]).isKnown()) {
            tw.writeVariableCell(UIColor::tc_Green, numToString(pl->getOreDensity(eles[i])) + "%");
        } else {
            tw.writeVariableCell(UIColor::tc_Green, string_t());
        }
    }
}
