/**
  *  \file client/widgets/dirchoose.cc
  *  \brief Game Directory Chooser Widget
  *
  *  (c) 2001-2011 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/widgets/dirchoose.h"
#include "cpluslib/format.h"
#include "gfx/font.h"
#include "io/io.h"
#include "util/nls.h"
#include "util/recode.h"

static int
getLevel(Ptr<Directory> dir)
{
    int level = 0;
    while (dir != 0 && level < 1000) {
        ++level;
        dir = dir->getParent();
    }
    return level;
}

/************************ WDirectoryChooser::Entry ***********************/

WDirectoryChooser::Entry::Entry(Action a, const string_t& name, Ptr<Directory> entry)
    : a(a), name(name), entry(entry)
{ }

WDirectoryChooser::Entry::~Entry()
{ }

/*************************** WDirectoryChooser ***************************/

/** \class WDirectoryChooser
    \brief Game Directory Chooser

    Widget that displays a directory list and lets the user choose from it. */

/** Initialize. Widget starts by displaying current working directory.
    \param id widget id. */
WDirectoryChooser::WDirectoryChooser(int id)
    : UIListbox(id, 14, 300, 0),
      current(Directory::openDir(Directory::getWorkingDirectory()))
{
    sig_dblclick.add(this, &WDirectoryChooser::enterCurrent);
    scan(scanGamelist, current);
}

/** Initialize.
    \param dir   initial directory
    \param oneup if true, starts by displaying directory's parent with
                 cursor on directory; if false, starts by displaying
                 directory's content
    \param id    widget id */
WDirectoryChooser::WDirectoryChooser(Ptr<Directory> dir, bool oneup, int id)
    : UIListbox(id, 14, 200, 0), current(0)
{
    sig_dblclick.add(this, &WDirectoryChooser::enterCurrent);

    Ptr<Directory> parent;
    if (oneup && (parent = dir->getParent()) != 0) {
        scan(scanGamelist, parent);
        showDirectory(dir);
    } else {
        scan(scanGamelist, dir);
    }
}

WDirectoryChooser::~WDirectoryChooser()
{ }

void
WDirectoryChooser::drawPart(GfxCanvas& can, int from, int to)
{
    GfxContext ctx;
    ctx.useCanvas(can);
    ctx.useColorScheme(getSkin());
    ctx.useFont(*fonts[FONT_NORMAL]);

    for (int line = from; line < to; ++line) {
        const int index = getTop() + line;
        const int y = extent.y + getItemHeight() * line;

        GfxRect area(extent.x, y, extent.w, getItemHeight());
        if (index >= int(content.size())) {
            drawBackground(ctx, area);
        } else {
            if (index == getCurrent()) {
                if (getFocusState() != NoFocus) {
                    drawSolidBar(ctx, area, UIColor::tc_Static);
                    ctx.setColor(UIColor::tc_InvStatic);
                } else {
                    drawBackground(ctx, area);
                    ctx.setColor(UIColor::tc_Static);
                    drawRectangle(ctx, area);
                }
            } else {
                drawBackground(ctx, area);
                ctx.setColor(UIColor::tc_Static);
            }
            outTextF(ctx, area.x + 10, area.y, area.w - 10, content[index].name);
        }
    }
}

/** Enter current directory. When something goes wrong, tries to recover
    (= stay in current directory). */
void
WDirectoryChooser::enterCurrent()
{
    if (getCurrent() < int(content.size())) {
        Ptr<Directory> c = current;
        try {
            switch (content[getCurrent()].a) {
             case aRescanWithoutGamelist:
                scan(scanNoGamelist, current);
                break;
             case aChdir:
             case aParent:
                scan(scanGamelist, content[getCurrent()].entry);
                break;
             case aShowRoots:
                scan(scanRoots, 0);
                break;
            }
        }
        catch(...) {
            /* no matter what happens, try to recover by returning
               to old directory. If that fails, bomb out. */
            scan(c ? scanGamelist : scanRoots, c);
        }
    }
}

/** Enter parent directory. */
void
WDirectoryChooser::enterParent()
{
    if (current != 0) {
        if (Ptr<Directory> p = current->getParent()) {
            scan(scanGamelist, p);
        }
    }
}

/** Enter a directory.
    \param dir Directory. May be null to show roots. */
void
WDirectoryChooser::enterDirectory(Ptr<Directory> dir)
{
    if (dir)
        scan(scanGamelist, dir);
    else
        scan(scanRoots, 0);
}

/** Show directory. Places the cursor on dir (or one of its parents)
    shown in the current display. */
void
WDirectoryChooser::showDirectory(Ptr<Directory> dir)
{
    // Skip if dir is null
    if (!dir) {
        setCurrent(0);
        return;
    }

    // Find both directories' levels
    int current_level = getLevel(current);
    int dir_level     = getLevel(dir);

    // Adjust so that dir is exactly beneath current
    ++current_level;
    while (dir_level > current_level) {
        dir = dir->getParent();
        --dir_level;
    }

    // Match?
    int n = 0;
    if (dir_level == current_level) {
        for (std::vector<Entry>::size_type i = 0; i != content.size(); ++i) {
            if (content[i].a == aChdir && content[i].entry->getName() == dir->getName()) {
                n = i;
                break;
            }
        }
    }
    setCurrent(n);
}

/** Get current subdirectory (=the directory the cursor is on).
    \return directory, or 0 if none (e.g. if cursor is on "[Parent]" entry */
Ptr<Directory>
WDirectoryChooser::getSubdirectory()
{
    if (getCurrent() < int(content.size()) && content[getCurrent()].a == aChdir)
        return content[getCurrent()].entry;
    return 0;
}

/** Initialize directory. This will read the directory, and update display
    on success.
    \param arg    modifies reading parameters
    \param dir    directory to read */
void
WDirectoryChooser::scan(ScanArg arg, Ptr<Directory> dir)
{
    content.clear();

    if (arg == scanRoots) {
        /* scan roots (=devices) */
        Ptr<Enumeration<Ptr<Directory> > > roots = Directory::getRoots();
        while (roots->hasMoreElements()) {
            Ptr<Directory> dir = roots->getNextElement();
            content.push_back(Entry(aChdir, dir->getTitle(), dir));
        }
    } else {
        if (Ptr<Directory> p = dir->getParent())
            content.push_back(Entry(aParent, _("[Up one level]"), p));

        if (arg == scanGamelist && readGamelist(*dir)) {
            /* ok */
        } else {
            /* no game list requested or none found */
            scanDirectory(*dir);
        }

        content.push_back(Entry(aShowRoots, _("[Other Places]"), 0));
    }

    Ptr<Directory> previous = current;
    current = dir;
    setSize(content.size());
    showDirectory(previous);
    adjustPosition();
    drawWidget(false);

    sig_dirchange.raise();
    sig_moved.raise(getId());
}

/** Scan directory content. */
void
WDirectoryChooser::scanDirectory(Directory& dir)
{
    std::vector<Entry>::size_type first_index = content.size();
    Ptr<Enumeration<Ptr<DirectoryEntry> > > list = dir.getEntries();
    while (list->hasMoreElements()) {
        Ptr<DirectoryEntry> entry = list->getNextElement();
        if (entry->getFileType() == DirectoryEntry::FT_DIRECTORY)
            content.push_back(Entry(aChdir, entry->getBaseName(),
                                    entry->openDir()));
    }
    std::sort(content.begin() + first_index, content.end(), isEntryLT);
}

/** Read gamestat.dat file. */
bool
WDirectoryChooser::readGamelist(Directory& dir)
{
    // Open file
    Ptr<Stream> s = dir.openFileNT("gamestat.dat", Stream::C_READ);
    if (!s)
        return false;

    // Read content
    char buffer[8][206];
    if (s->read(buffer[0], sizeof buffer) != sizeof buffer)
        return false;

    // Check content
    for (int i = 0; i < 8; ++i) {
        content.push_back(Entry(aChdir, " " + convertGameToUtf8(getBasicStringN(&buffer[i][184], 20)),
                                dir.openSubdir(format("vpwork%d", i+1))));
    }
    content.push_back(Entry(aRescanWithoutGamelist, _("[Directory Content]"), 0));
    return true;
}

/** Compare two entries. */
bool
WDirectoryChooser::isEntryLT(const Entry& l, const Entry& r)
{
    return strCaseCmp(l.name, r.name) < 0;
}
