/**
  *  \file client/wtile.cc
  *
  *  (c) 2001-2011 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/wtile.h"
#include "ui/ui.h"
#include "ui/uiglobal.h"
#include "game/obj.h"
#include "ui/layout.h"
#include "gfx/clip.h"

/********************************* WTile *********************************/

/** \class WTile
    \brief Widget displaying certain data (hooray.)

    Tiles can display the following stuff:
    - a constant heading
    - some fixed data
    - some variable data

    This class provides an interface intended to make display of such
    stuff easier. It automatically manages update events. In addition,
    a tile can request to have some associated controls (buttons).

    To avoid excess redraws, data is partitioned. Each tile knows on
    which events it wants to be updated (\c what parameter to
    constructor). Update events (see update(int)) include this
    parameter; the tile is redrawn if any of the bits in our \c what
    is set in the event, or if \c what is zero (meaning that we always
    want to be updated) or the event is zero (meaning that we don't
    know what was updated and thus update everything just-in-case).

    \todo I'm not 100% sure whether the redraw optimisations are still
    needed or whether we can simply update everything.

    (In PCC 1.x, the distinguishing property of a "Tile" was that it
    was freely moveable. In PCC 2, all widgets are moveable.) */

/** Standard constructor.
    \param size  size of this tile
    \param id    widget Id of this tile */
WTile::WTile(GfxPoint size, int id)
    : UIBaseComplexWidget(id),
      size(size),
      children(UINullLayout::instance),
      state(sContentOnly),
      layout(LeftJustified),
      lrmargin(0),
      children_move(false)
{
    addChildWidget(children, 0);
    computeCoordinates();
}

WTile::~WTile()
{ }

/** Draw the tile. Do not override in WTile descendants. */
void
WTile::drawContent(GfxCanvas& can)
{
    if (state != sContentOnly) {
        GfxContext ctx;
        ctx.useCanvas(can);
        ctx.useColorScheme(getSkin());
        ctx.useFont(*fonts[FONT_TITLE]);
        ctx.setColor(UIColor::tc_Heading);
        outTextF(ctx, extent.x, extent.y, extent.w, heading);
    }
    if (state != sHeadingOnly) {
        GfxOffset ofs(extent.x, extent.y + y_origin, can);
        drawData(ofs);
    }
    can.update(getExtent());
}

/** Set tile layout. All child controls of the tile are arranged in a bounding box.
    Depending on the alignment, this box is pushed to the left or right side of the tile,
    and is either aligned with the content (and thus moves if the heading is switched on
    or off) or with the tile.
    \param layout   Alignment
    \param lrmargin Room to leave from left/right edge of tile, in pixels
    \param children_move true if children move with content, false if they are fixed relative to top of tile */
void
WTile::setLayout(Layout layout, int lrmargin, bool children_move)
{
    this->layout        = layout;
    this->lrmargin      = lrmargin;
    this->children_move = children_move;
}

/** Add control to tile.
    \param r Bounding rectangle relative to origin defined with setLayout(), see there.
    \param w Widget to add */
void
WTile::addControl(GfxRect r, UIBaseWidget& w)
{
    // coordinates are relative to children group
    r.moveBy(children.getExtent().getTopLeft());

    // add it
    w.setExtent(r);
    children.add(w);
}

void
WTile::onResize(GfxRect /*oldsize*/, bool /*by_layout*/)
{
    /* Compute bounding box of children. Relative point (0,0) is always included.
       This makes tiles work that look like this:
                        ,-----------------------------.
                        |         .------------------,|
         Bounding box ------------->             .--.||
                        |         |              |  <--- Widget
               Tile ------>       |              `--'||
                        |         `------------------'|
                        `-----------------------------'
       Those are created by having the leftmost widget not start at X=0. */
    GfxRect bbox(children.getExtent().getTopLeft(), GfxPoint(1, 1));
    for (UIBaseWidget* p = children.getFirstChild(); p != 0; p = p->getNextSibling())
        bbox.include(p->getExtent());

    /* Compute new position */
    GfxRect mysize = getExtent();
    if (layout == LeftJustified) {
        bbox.w += mysize.x - bbox.x;
        bbox.x =  mysize.x + lrmargin;
    } else {
        bbox.x = mysize.x + mysize.w - bbox.w - lrmargin;
    }
    bbox.y = mysize.y;
    if (children_move)
        bbox.y += y_origin;
    
    children.setExtent(bbox, true);
}

/** Get Layout Information. Do not override in WTile descendants. */
void
WTile::getLayoutInfo(LayoutInfo& info)
{
    GfxPoint size = this->size;
    if (state == sHeadingOnly)
        size.y = 0;
    size.y += y_origin;

    info.flags = 0;
    info.min_size = info.pref_size = size;
}

/** Get size of data area, with margins. This returns the dimensions
    of the data area you are expected to fill, in relative coordinates.
    \param lmargin left margin (number of pixels to remove at the left)
    \param rmargin right margin */
GfxRect
WTile::getInnerSize(int lmargin, int rmargin)
{
    GfxRect rv = getExtent();
    rv.x  = lmargin;
    rv.w -= lmargin+rmargin;
    rv.y  = 0;
    rv.h -= y_origin;
    return rv;
}

/** Set display state.
    \todo Currently, this only works reliably when the tile is not
    yet in a group. It does not actually resize the tile. */
void
WTile::setDisplayState(State state)
{
    if (state == this->state)
        return;
    this->state = state;
    computeCoordinates();
    onResize(getExtent(), true);
    drawWidget();
    // FIXME: if widget is *really* already displayed, we probably
    // should tell our parent to resize us
}

/** Set heading of the tile.
    \return *this, for convenience */
WTile&
WTile::setHeading(string_t heading)
{
    this->heading = heading;
    drawWidget();
    return *this;
}

/** Compute coordinates, after a change to layout parameters. \internal */
void
WTile::computeCoordinates()
{
    if (state == sContentOnly) {
        y_origin = 0;
    } else {
        y_origin = font_heights[FONT_TITLE];
    }
}

/************************ WObjectTypeObserverTile ************************/

/*! \class WObjectTypeObserverTile
    \brief Tile observing an GObjectType's selected GObject

    The WObjectTypeObserverTile observes a GObjectType. Whenever a new object
    is selected on the GObjectType, or the selected object changes, the tile
    is redrawn. */

/** Constructor.
    \param size Size of tile
    \param type GObjectType to observe */
WObjectTypeObserverTile::WObjectTypeObserverTile(GfxPoint size, GObjectType& type)
    : WTile(size),
      GObjectTypeObserver(type)
{ }

WObjectTypeObserverTile::~WObjectTypeObserverTile()
{ }

/** Called when the selected object changes. */
void
WObjectTypeObserverTile::onObjectChanged()
{
    drawWidget();
}
