/**
  *  \file game/parser-name.cc
  *
  *  (c) 2001-2011 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "game/parser.h"
#include "util/nls.h"
#include "cpluslib/misc.h"

const char*
getNameForItem(GMessageStringIndex si)
{
    switch (si) {
     case ms_Name:  return _("Name");
     case ms_FCode: return _("FCode");
     case ms_Max:   return "?";
    }
    return "?";
}

const char*
getNameForItem(GMessageIntIndex ii)
{
    switch (ii) {
     case mi_X:                 return _("X");
     case mi_Y:                 return _("Y");
     case mi_Radius:            return _("Radius");
     case mi_Owner:             return _("Owner");
     case mi_Damage:            return _("Damage");
     case mi_Heading:           return _("Heading");
     case mi_Speed:             return _("Speed");

        /* Minefields: */
     case mi_MineUnits:         return _("Mine Units");
     case mi_MineScanReason:    return _("Mine Scan Reason");
     case mi_MineType:          return _("Mine Type");
     case mi_MineUnitsRemoved:  return _("Mine Units Removed");

        /* Ships: */
     case mi_ShipHull:          return _("Hull Type");
     case mi_ShipFuel:          return _("Fuel");

        /* Planets: */
     case mi_PlanetTotalN:      return _("Total Neutronium");
     case mi_PlanetTotalT:      return _("Total Tritanium");
     case mi_PlanetTotalD:      return _("Total Duranium");
     case mi_PlanetTotalM:      return _("Total Molybdenum");
     case mi_PlanetAddedN:      return _("Added Neutronium");
     case mi_PlanetAddedT:      return _("Added Tritanium");
     case mi_PlanetAddedD:      return _("Added Duranium");
     case mi_PlanetAddedM:      return _("Added Molybdenum");
     case mi_PlanetMinedN:      return _("Mined Neutronium");
     case mi_PlanetMinedT:      return _("Mined Tritanium");
     case mi_PlanetMinedD:      return _("Mined Duranium");
     case mi_PlanetMinedM:      return _("Mined Molybdenum");
     case mi_PlanetCash:        return _("Money");
     case mi_PlanetSupplies:    return _("Supplies");
     case mi_PlanetHasBase:     return _("Has Base");
     case mi_PlanetMines:       return _("Mineral Mines");
     case mi_PlanetFactories:   return _("Factories");
     case mi_PlanetDefense:     return _("Defense Posts");
     case mi_PlanetTemperature: return _("Temperature");
     case mi_PlanetColonists:   return _("Colonist Clans");
     case mi_PlanetActivity:    return _("Industrial activity");
     case mi_PlanetNativeRace:  return _("Native Race");
     case mi_PlanetNativeGov:   return _("Native Government");
     case mi_PlanetNatives:     return _("Native Population");
     case mi_PlanetHasNatives:  return _("Natives present");

        /* Bases: */
     case mi_BaseQueuePos:      return _("Build Queue Position");
     case mi_BaseQueuePriority: return _("Build Priority");

        /* Score: */
     case mi_ScoreWinLimit:     return _("Score Win Limit");
     case mi_ScoreTurnLimit:    return _("Score Turn Limit");

        /* Ion Storm: */
     case mi_IonVoltage:        return _("Voltage");
     case mi_IonStatus:         return _("Storm Status");


     case mi_Max:               return "?";
    }
    return "?";
}

namespace {
    struct Keyword {
        uint16_t index;
        const char* word;
    };
}

static const Keyword stringNames[] = {
    { ms_FCode,             "FCODE"          },
    { ms_Name,              "NAME"           },
};

static const Keyword intNames[] = {
    { mi_PlanetAddedD,      "ADDED.D"        },
    { mi_PlanetAddedM,      "ADDED.M"        },
    { mi_PlanetAddedN,      "ADDED.N"        },
    { mi_PlanetAddedT,      "ADDED.T"        },
    { mi_PlanetHasBase,     "BASE"           },
    { mi_BaseQueuePriority, "BUILD.PRIORITY" }, // Not in scripting language
    { mi_BaseQueuePos,      "BUILD.QPOS"     },
    { mi_ShipFuel,          "CARGO.N"        },
    { mi_PlanetColonists,   "COLONISTS"      },
    { mi_Damage,            "DAMAGE"         },
    { mi_PlanetDefense,     "DEFENSE"        },
    { mi_PlanetFactories,   "FACTORIES"      },
    { mi_Heading,           "HEADING"        },
    { mi_ShipHull,          "HULL"           },
    { mi_PlanetActivity,    "INDUSTRY"       },
    { mi_PlanetMinedD,      "MINED.D"        },
    { mi_PlanetMinedM,      "MINED.M"        },
    { mi_PlanetMinedN,      "MINED.N"        },
    { mi_PlanetMinedT,      "MINED.T"        },
    { mi_PlanetMines,       "MINES"          },
    { mi_PlanetCash,        "MONEY"          },
    { mi_PlanetNatives,     "NATIVES"        },
    { mi_PlanetNativeGov,   "NATIVES.GOV"    },
    { mi_PlanetNativeRace,  "NATIVES.RACE"   },
    { mi_PlanetHasNatives,  "NATIVES.YESNO"  }, // Not in scripting language
    { mi_Owner,             "OWNER"          },
    { mi_Radius,            "RADIUS"         },
    { mi_MineScanReason,    "REASON"         }, // SCANNED in the scripting language
    { mi_Speed,             "SPEED"          },
    { mi_IonStatus,         "STATUS"         },
    { mi_PlanetSupplies,    "SUPPLIES"       },
    { mi_PlanetTemperature, "TEMP"           },
    { mi_PlanetTotalD,      "TOTAL.D"        },
    { mi_PlanetTotalM,      "TOTAL.M"        },
    { mi_PlanetTotalN,      "TOTAL.N"        },
    { mi_PlanetTotalT,      "TOTAL.T"        },
    { mi_ScoreTurnLimit,    "TURNLIMIT"      },
    { mi_MineType,          "TYPE"           },
    { mi_MineUnits,         "UNITS"          },
    { mi_MineUnitsRemoved,  "UNITS.REMOVED"  }, // Not in scripting language
    { mi_IonVoltage,        "VOLTAGE"        },
    { mi_ScoreWinLimit,     "WINLIMIT"       },
    { mi_X,                 "X"              },
    { mi_Y,                 "Y"              },
};

static int
lookupKeyword(const string_t& kw, const Keyword defs[], std::size_t ndefs, int dflt)
{
    // Slow and simple
    for (std::size_t i = 0; i < ndefs; ++i)
        if (defs[i].word == kw)
            return defs[i].index;
    return dflt;
}

GMessageStringIndex
getStringIndexFromKeyword(string_t kw)
{
    return GMessageStringIndex(lookupKeyword(kw, stringNames, countof(stringNames), ms_Max));
}

GMessageIntIndex
getIntIndexFromKeyword(string_t kw)
{
    return GMessageIntIndex(lookupKeyword(kw, intNames, countof(intNames), mi_Max));
}
