/**
  *  \file int/globalif.cc
  *  \brief Interpreter Game Interface: Global Functions and Properties
  *
  *  (c) 2001-2011 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <memory>
#include <cmath>
#include <ctime>
#include <limits.h>
#include "int/if/globalif.h"
#include "game/coord.h"
#include "int/data.h"
#include "int/value.h"
#include "int/compound.h"
#include "int/context.h"
#include "int/interror.h"
#include "int/if/ifutil.h"
#include "game/struct.h"
#include "game/global.h"
#include "game/planet.h"
#include "game/ship.h"
#include "game/config.h"
#include "game/team.h"
#include "version.h"
#include "game/host.h"
#include "cpluslib/misc.h"
#include "int/intglobal.h"
#include "game/fcode.h"
#include "int/if/playerif.h"
#include "game/cmds.h"
#include "cpluslib/format.h"
#include "util/nls.h"
#include "game/reg.h"
#include "game/cspec.h"
#include "game/drawing.h"
#include "int/args.h"
#include "client/marks.h"
#include "util/random.h"
#include "game/sel.h"
#include "int/if/userif.h"
#include "int/contextio.h"
#include "game/gamedirs.h"
#include "game/storage.h"
#include "game/selio.h"
#include "util/answer.h"
#include "int/file.h"
#include "int/expr.h"
#include "int/exec.h"
#include "int/process.h"

class TextFile;

namespace {
    enum { GlobalProperty, MyPlayerProperty, UIProperty };
}

static const TIntMapping global_mapping[] = {
    { "CHART.X",               iuiChartX,            UIProperty },
    { "CHART.Y",               iuiChartY,            UIProperty },
    { "MY.BASES",              iplScoreBases,        MyPlayerProperty },
    { "MY.INMSGS",             igpMyInMsgs,          GlobalProperty },
    { "MY.OUTMSGS",            igpMyOutMsgs,         GlobalProperty },
    // MY.PBPS
    { "MY.PLANETS",            iplScorePlanets,      MyPlayerProperty },
    { "MY.RACE",               iplShortName,         MyPlayerProperty },
    { "MY.RACE$",              iplId,                MyPlayerProperty },
    { "MY.RACE.ADJ",           iplAdjName,           MyPlayerProperty },
    { "MY.RACE.FULL",          iplFullName,          MyPlayerProperty },
    { "MY.RACE.ID",            iplRaceId,            MyPlayerProperty },
    { "MY.RACE.MISSION",       iplMission,           MyPlayerProperty },
    { "MY.SCORE",              iplScore,             MyPlayerProperty },
    { "MY.SHIPS",              iplScoreShips,        MyPlayerProperty },
    { "MY.SHIPS.CAPITAL",      iplScoreCapital,      MyPlayerProperty },
    { "MY.SHIPS.FREIGHTERS",   iplScoreFreighters,   MyPlayerProperty },
    { "MY.TEAM",               iplTeam,              MyPlayerProperty },
    { "MY.VCRS",               igpMyVCRs,            GlobalProperty },
    { "SELECTION.LAYER",       igpSelectionLayer,    GlobalProperty },
    { "SHIPS.CAPITAL",         iplTotalCapital,      MyPlayerProperty },  // Implemented as player properties because
    { "SHIPS.FREIGHTERS",      iplTotalFreighters,   MyPlayerProperty },  // that has easier access to a 'TGen'. A better
    { "SHIPS.TOTAL",           iplTotalShips,        MyPlayerProperty },  // way may be to implement it using GStatFile.
    // SYSTEM.EGG
    // SYSTEM.ERR -> global variable (localizable!)
    { "SYSTEM.GAMEDIRECTORY",  igpGameDirectory,     GlobalProperty },
    { "SYSTEM.GAMETYPE",       igpRegSharewareText,  GlobalProperty },
    { "SYSTEM.GAMETYPE$",      igpRegSharewareFlag,  GlobalProperty },
    // SYSTEM.GUI -> conif, guiif
    { "SYSTEM.HOST",           igpSystemHost,        GlobalProperty },
    { "SYSTEM.HOST$",          igpSystemHostCode,    GlobalProperty },
    { "SYSTEM.HOSTVERSION",    igpSystemHostVersion, GlobalProperty },
    { "SYSTEM.LOCAL",          igpFileFormatLocal,   GlobalProperty },
    { "SYSTEM.PROGRAM",        igpSystemProgram,     GlobalProperty },
    { "SYSTEM.RANDOMSEED",     igpRandomSeed,        GlobalProperty },
    { "SYSTEM.REGSTR1",        igpRegStr1,           GlobalProperty },
    { "SYSTEM.REGSTR2",        igpRegStr2,           GlobalProperty },
    { "SYSTEM.REMOTE",         igpFileFormatRemote,  GlobalProperty },
    { "SYSTEM.ROOTDIRECTORY",  igpRootDirectory,     GlobalProperty },
    { "SYSTEM.SIM",            iuiSimFlag,           UIProperty },
    { "SYSTEM.VERSION",        igpSystemVersion,     GlobalProperty },
    { "SYSTEM.VERSION$",       igpSystemVersionCode, GlobalProperty },
    { "TURN",                  igpTurnNumber,        GlobalProperty },
    { "TURN.DATE",             igpTurnDate,          GlobalProperty },
    { "TURN.ISNEW",            igpTurnIsNew,         GlobalProperty },
    { "TURN.TIME",             igpTurnTime,          GlobalProperty },
    { "UI.SCREEN",             iuiScreenNumber,      UIProperty },
    { "UI.X",                  iuiScanX,             UIProperty },
    { "UI.Y",                  iuiScanY,             UIProperty },
};

static RandomNumberGenerator script_rng(std::time(0));

/**************************** IntGlobalContext ***************************/

IntGlobalContext::IntGlobalContext()
{ }

IntGlobalContext::~IntGlobalContext()
{ }

int32_t
IntGlobalContext::lookup(string_t name)
{
    static const uint32_t nproperties = countof(global_mapping);

    if (name.compare(0, 7, "GLOBAL.", 7) == 0)
        name.erase(0, 7);

    int32_t result = global_property_names.lookup(name);
    if (result >= 0)
        return result + nproperties;

    return lookupName(name, global_mapping, countof(global_mapping));
}

void
IntGlobalContext::set(int32_t index, IntValue* value)
{
    static const int32_t nproperties = countof(global_mapping);
    if (index >= nproperties) {
        /* User variable */
        global_values.set(index - nproperties, value);
    } else {
        /* Global property */
        switch (global_mapping[index].domain) {
         case GlobalProperty:
            setGlobalProperty(IntGlobalProperty(global_mapping[index].index), value);
            break;
         case UIProperty:
            IntUserInterfaceBinding::getCurrentInstance()->setProperty(IntUserInterfaceProperty(global_mapping[index].index), value);
            break;
         default:
            throw IntError::notAssignable();
        }
    }
}

IntValue*
IntGlobalContext::get(int32_t index)
{
    /* Check user variables */
    static const int32_t nproperties = countof(global_mapping);
    if (index >= nproperties) {
        return IntValue::cloneOf(global_values[index - nproperties]);
    }

    /* Check global properties */
    int n;
    switch (global_mapping[index].domain) {
     case GlobalProperty:
        return getGlobalProperty(IntGlobalProperty(global_mapping[index].index));
     case MyPlayerProperty:
        return getPlayerProperty(getPlayerId(), IntPlayerProperty(global_mapping[index].index));
     case UIProperty:
        n = IntUserInterfaceBinding::getCurrentInstance()->getProperty(IntUserInterfaceProperty(global_mapping[index].index));
        if (n >= 0) {
            return makeIntValue(n);
        } else {
            return 0;
        }
    }
    return 0;
}

IntContext*
IntGlobalContext::clone()
{
    return new IntGlobalContext();
}

GObject*
IntGlobalContext::getObject()
{
    return 0;
}

string_t
IntGlobalContext::toString(bool /*readable*/)
{
    return "#<global>";
}

void
IntGlobalContext::store(IntTagNode& sv, Stream& /*aux*/)
{
    sv.tag   = Tag_Global;
    sv.value = 0;
}



/** Get global property.
    \param igp Global property to query. */
IntValue*
getGlobalProperty(IntGlobalProperty igp)
{
    switch (igp) {
     case igpFileFormatLocal:
        /* In PCC 1.x, this value is 1 for Dosplan, 2 for Winplan, and
           determines the format of control and message outbox files.
           In PCC2, those are orthogonal. We can, however, comparatively
           easily determine the mailbox format. */
        if (haveDisplayedTurn()) {
            if (getCurrentTurn().getOutbox(getPlayerId()).getFormat() == GOutbox::fDos)
                return makeIntValue(1);
            else
                return makeIntValue(2);
        } else {
            return 0;
        }
     case igpFileFormatRemote:
        /* In PCC 1.x, this value is 0 for Dosplan, 1 for Winplan, and
           determines the TRN format. This is the config option. As of
           20110206, PCC2 has no such config option and always produces
           Winplan format. */
        return makeIntValue(1);
     case igpGameDirectory:
        if (game_dir_name.size())
            return makeStringValue(game_dir_name);
        else
            return 0;
     case igpMyInMsgs:
        if (haveDisplayedTurn())
            return makeIntValue(getDisplayedTurn().getInbox().getCount());
        else
            return 0;
     case igpMyOutMsgs:
        if (haveDisplayedTurn())
            return makeIntValue(getDisplayedTurn().getOutbox(getPlayerId()).getCount());
        else
            return 0;
     case igpMyVCRs:
        if (haveDisplayedTurn())
            return makeIntValue(getDisplayedTurn().getVcrs().getCount());
        else
            return 0;
     case igpRootDirectory:
        if (root_dir_name.size())
            return makeStringValue(root_dir_name);
        else
            return 0;
     case igpSelectionLayer:
        return makeIntValue(GMultiSelection::getCurrentSelectionLayer());
     case igpSystemProgram:
        return makeStringValue("PCC");
     case igpSystemVersion:
        return makeStringValue(PCC2_VERSION);
     case igpSystemVersionCode:
        return makeIntValue(PCC2_VERSION_CODE);
     case igpSystemHost:
        if (host.isPHost())
            return makeStringValue("PHost");
        else if (host.isSRace())
            return makeStringValue("SRace");
        else if (host.isTHost())
            return makeStringValue("Host");
        else
            return 0;
     case igpSystemHostCode:
        if (host.isPHost())
            return makeIntValue(2);
        else if (host.isSRace())
            return makeIntValue(1);
        else if (host.isTHost())
            return makeIntValue(0);
        else
            return 0;
     case igpSystemHostVersion:
        return makeIntValue(host.host_version);
     case igpRegSharewareFlag:
        return makeBoolValue(!getRegInfo().isRegistered());
     case igpRegSharewareText:
        return makeStringValue(getRegInfo().isRegistered() ? "Registered" : "Shareware");
     case igpRandomSeed:
        return makeIntValue(script_rng.getSeed());
     case igpRegStr1:
        return makeStringValue(getRegInfo().getRegStr1());
     case igpRegStr2:
        return makeStringValue(getRegInfo().getRegStr2());
     case igpTurnNumber:
        if (haveDisplayedTurn())
            return makeIntValue(getDisplayedTurn().getTurnNumber());
        else
            return 0;
     case igpTurnIsNew:
        if (haveDisplayedTurn())
            return makeBoolValue(getDisplayedTurn().getDatabaseTurn() < getDisplayedTurn().getTurnNumber());
        else
            return 0;
     case igpTurnTime:
        if (haveDisplayedTurn())
            return makeStringValue(getDisplayedTurn().getCurrentUniverse().getTimestamp().getTimeAsString());
        else
            return 0;
     case igpTurnDate:
        if (haveDisplayedTurn())
            return makeStringValue(getDisplayedTurn().getCurrentUniverse().getTimestamp().getDateAsString());
        else
            return 0;
    }
    return 0;
}

/** Set global property.
    \param igp Global property to set
    \param value New value */
void
setGlobalProperty(IntGlobalProperty igp, IntValue* value)
{
    int32_t iv;
    switch (igp) {
     case igpSelectionLayer:
        if (checkIntArg(iv, value, 0, GMultiSelection::NUM_SELECTION_LAYERS-1))
            GMultiSelection::setCurrentSelectionLayer(iv);
        break;
     case igpRandomSeed:
        if (checkIntArg(iv, value))
            script_rng.setSeed(iv);
        break;
     default:
        throw IntError::notAssignable();
    }
}

/******************************* Functions *******************************/

/** Implementation of the "Distance" function. */
IntValue*
IFDistanceGet(IntArgBlock& args)
{
    GPoint points[2];

    for (int i = 0; i < 2; ++i) {
        /* Do we have an argument? */
        if (args.getNumArgs() == 0)
            throw IntError("Too few arguments for \"Distance\"");
        IntValue* theValue = args.getNext();

        /* What is it? */
        if (theValue == 0) {
            /* Null. Result is null. */
            return 0;
        } else if (IntContext* cv = dynamic_cast<IntContext*>(theValue)) {
            /* Context. Must have LOC.X and LOC.Y properties. */
            int32_t x = cv->lookup("LOC.X");
            int32_t y = cv->lookup("LOC.Y");
            if (x < 0 || y < 0)
                throw IntError("Operand doesn't have a position");

            /* Fetch values */
            std::auto_ptr<IntValue> vx(cv->get(x));
            std::auto_ptr<IntValue> vy(cv->get(y));
            if (!vx.get() || !vy.get())
                return 0;

            /* Must be integers */
            IntIntValue* vix = dynamic_cast<IntIntValue*>(vx.get());
            IntIntValue* viy = dynamic_cast<IntIntValue*>(vy.get());
            if (!vix || !viy)
                throw IntError::typeError(IntError::tyInteger);

            points[i].x = vix->getValue();
            points[i].y = viy->getValue();
        } else if (IntIntValue* iv = dynamic_cast<IntIntValue*>(theValue)) {
            /* Integer. There must be another integer. */
            if (args.getNumArgs() == 0)
                throw IntError("Too few arguments for \"Distance\"");
            IntValue* otherValue = args.getNext();

            /* Null? */
            if (!otherValue)
                return 0;

            /* Integer? */
            IntIntValue* iv2 = dynamic_cast<IntIntValue*>(otherValue);
            if (!iv2)
                throw IntError::typeError(IntError::tyInteger);

            points[i].x = iv->getValue();
            points[i].y = iv2->getValue();
        } else {
            /* Error. */
            throw IntError::typeError();
        }
    }

    if (args.getNumArgs() != 0)
        throw IntError("Too many arguments to \"Distance\"");

    return makeFloatValue(std::sqrt(double(distanceSquared(points[0], points[1]))));
}

/** Implementation of the "ShipName" function. */
IntValue*
IFShipNameGet(IntArgBlock& args)
{
    int32_t sid;

    args.checkArgumentCount(1);
    if (!checkIntArg(sid, args.getNext()))
        return 0;

    if (sid <= 0 || sid > NUM_SHIPS)
        return 0;

    if (!haveDisplayedTurn())
        return 0;

    return makeStringValue(getDisplayedTurn().getCurrentUniverse().getShip(sid).getName());
}

/** Implementation of the "PlanetName" function. */
IntValue*
IFPlanetNameGet(IntArgBlock& args)
{
    int32_t sid;

    args.checkArgumentCount(1);
    if (!checkIntArg(sid, args.getNext()))
        return 0;

    if (sid <= 0 || sid > NUM_PLANETS)
        return 0;

    if (!haveDisplayedTurn())
        return 0;

    return makeStringValue(getDisplayedTurn().getCurrentUniverse().getPlanet(sid).getName());
}

/** Implementation of the "PlanetAt" function. */
IntValue*
IFPlanetAtGet(IntArgBlock& args)
{
    int32_t x, y;

    // Fetch x,y parameters
    args.checkArgumentCount(2, 3);
    if (!checkIntArg(x, args.getNext(), -32767, 32767) || !checkIntArg(y, args.getNext(), -32767, 32767))
        return 0;

    // Fetch optional flag argument
    bool flag = false;
    if (args.getNumArgs() > 0) {
        // FIXME: this is inconsistent in that a missing argument does not equal a null argument
        int n = getBoolValue(args.getNext());
        if (n < 0)
            return 0;
        flag = (n != 0);
    }

    // Generate result
    if (!haveDisplayedTurn())
        return 0;

    return makeIntValue(getDisplayedTurn().getCurrentUniverse().getPlanetAt(GPoint(x, y), flag));
}

/** Implementation of the "Cfg" function. */
IntValue*
IFCfgGet(IntArgBlock& args)
{
    args.checkArgumentCount(1, 2);

    /* Config key */
    string_t optName;
    if (!checkStringArg(optName, args.getNext()))
        return 0;
    ConfigOption* opt = config.getConfigOption(optName);
    if (opt == 0)
        throw IntError("Invalid first argument to \"Cfg\"");

    /* Player number */
    int32_t player = 0;
    if (args.getNumArgs() > 0)
        if (!checkIntArg(player, args.getNext(), 1, NUM_PLAYERS))
            throw IntError("Invalid second argument to \"Cfg\"");

    /* Figure out option */
    if (ConfigAlias* a = dynamic_cast<ConfigAlias*>(opt))
        opt = &a->getForwardedOption();

    /* FIXME: ugly type switch. Can we do better? */
    if (ConfigIntArrayBaseOption* bopt = dynamic_cast<ConfigIntArrayBaseOption*>(opt)) {
        /* ints, optional player */
        if (!player) {
            /* Possible limits are
                 2    NewNativesPopulationRange
                 4    WraparoundRectangle
                 8    MeteorShowerOreRanges
                 9    NewNativesRaceRate
                 10   ConfigExpOption, e.g. EModBayRechargeRate
                 11   ConfigStdOption, e.g. RaceMiningRate */
            if (bopt->getLimit() == 11)
                player = getPlayerId();
            else
                throw IntError("Too few arguments");
        }
        if (player > bopt->getLimit())
            throw IntError::rangeError();
        return makeIntValue(bopt->getArray()[player-1]);
    } else if (ConfigIntOption* intopt = dynamic_cast<ConfigIntOption*>(opt)) {
        /* single int, no player. Example: NumShips */
        if (player)
            throw IntError("Too many arguments");
        return makeIntValue((*intopt)());
    } else if (ConfigCostOption* costopt = dynamic_cast<ConfigCostOption*>(opt)) {
        /* 11 costs. Example: StarbaseCost */
        if (!player)
            player = getPlayerId();
        return makeStringValue((*costopt)(player).toCargospec());
    } else if (ConfigStringOption* stropt = dynamic_cast<ConfigStringOption*>(opt)) {
        /* single string. FIXME: PCC 1.x splits ExperienceLevelNames */
        if (player)
            throw IntError("Too many arguments");
        return makeStringValue((*stropt)());
    } else {
        /* what? */
        return 0;
    }
}

/** Implementation of the "IsSpecialFCode" function. */
IntValue*
IFIsSpecialFCodeGet(IntArgBlock& args)
{
    /* Check arguments */
    args.checkArgumentCount(1);
    IntValue* arg = args.getNext();
    if (!arg)
        return 0;

    /* Function implementation */
    string_t fc = arg->toString(false).substr(0, 3);
    return makeBoolValue(GFCode::isSpecial(fc, true));
}

/** Implementation of the "Truehull" function. */
IntValue*
IFTruehull(IntArgBlock& args)
{
    int32_t slot, player;

    /* Check arguments */
    args.checkArgumentCount(1, 2);
    if (!checkIntArg(slot, args.getNext()))
        return 0;

    if (args.getNumArgs() > 0) {
        // FIXME: this is inconsistent in that a missing argument does not equal a null argument
        if (!checkIntArg(player, args.getNext()))
            return 0;
    } else {
        player = getPlayerId();
    }

    int result;
    if (slot <= 0 || slot > NUM_HULLS_PER_PLAYER || player <= 0 || player > NUM_PLAYERS)
        result = 0;
    else
        result = getTruehull(player, slot);

    return makeIntValue(result);
}

/** Implementation of the "RandomFCode" function. */
IntValue*
IFRandomFCode(IntArgBlock& args)
{
    args.checkArgumentCount(0);
    return makeStringValue(GFCode::generateRandomFCode());
}

/** Implementation of the "Format" function. */
IntValue*
IFFormat(IntArgBlock& args)
{
    enum { LIMIT = 10 };
    args.checkArgumentCount(1, LIMIT+1);

    // First, find the format string
    string_t fmt;
    if (!checkStringArg(fmt, args.getNext()))
        return 0;

    // The formatter is intended to be used as a temporary, and keeps references
    // to its arguments. So we must make sure they live sufficiently long.
    int32_t ints[LIMIT];
    double floats[LIMIT];
    string_t strings[LIMIT];
    format formatter(fmt.c_str());

    for (uint32_t i = 0, limit = args.getNumArgs(); i < limit; ++i) {
        // Check and convert value
        IntValue* v = args.getNext();
        if (v == 0)
            return 0;

        if (IntIntValue* iv = dynamic_cast<IntIntValue*>(v)) {
            ints[i] = iv->getValue();
            formatter << ints[i];
        } else if (IntFloatValue* fv = dynamic_cast<IntFloatValue*>(v)) {
            floats[i] = fv->getValue();
            formatter << floats[i];
        } else if (IntStringValue* sv = dynamic_cast<IntStringValue*>(v)) {
            strings[i] = sv->getValue();
            formatter << strings[i];
        } else {
            throw IntError::typeError(IntError::tyBaseType);
        }
    }

    // Format
    return makeStringValue(formatter);
}

/** Implementation of the "Translate" function. */
IntValue*
IFTranslate(IntArgBlock& args)
{
    args.checkArgumentCount(1);
    string_t s;
    if (!checkStringArg(s, args.getNext()))
        return 0;
    return makeStringValue(_(s.c_str()));
}

/** Implementation of the "Random" function. */
IntValue*
IFRandom(IntArgBlock& args)
{
    int32_t lo, hi;
    args.checkArgumentCount(1, 2);
    if (!checkIntArg(lo, args.getNext(), 0, 0x7FFF))
        return 0;

    if (args.getNumArgs() > 0) {
        if (!checkIntArg(hi, args.getNext(), 0, 0x7FFF))
            return 0;
    } else {
        hi = lo;
        lo = 0;
    }

    if (lo < hi) {
        return makeIntValue(lo + script_rng(hi - lo));
    } else if (lo > hi) {
        return makeIntValue(lo - script_rng(lo - hi));
    } else {
        return makeIntValue(lo);
    }
}

/** Implementation of the "CAdd" function. */
IntValue*
IFCAdd(IntArgBlock& args)
{
    // At least one arg (PCC 1.x: at least two)
    args.checkArgumentCountAtLeast(1);

    GCargoSpec sum;
    while (args.getNumArgs() != 0) {
        GCargoSpec summand;
        if (!checkCargoSpecArg(summand, args.getNext()))
            return 0;
        sum += summand;
    }

    return makeStringValue(sum.toCargoSpecString());
}

/** Implementation of the "CCompare" function. */
IntValue*
IFCCompare(IntArgBlock& args)
{
    args.checkArgumentCount(2);

    GCargoSpec a, b;
    if (!checkCargoSpecArg(a, args.getNext()))
        return 0;
    if (!checkCargoSpecArg(b, args.getNext()))
        return 0;

    return makeBoolValue(a.isEnoughFor(b));
}

/** Implementation of the "CDiv" function. */
IntValue*
IFCDiv(IntArgBlock& args)
{
    args.checkArgumentCount(2);

    GCargoSpec a;
    if (!checkCargoSpecArg(a, args.getNext()))
        return 0;

    IntValue* arg = args.getNext();
    if (arg == 0) {
        // Null
        return 0;
    } else if (IntIntValue* iv = dynamic_cast<IntIntValue*>(arg)) {
        // Divide cargospec by integer
        if (!a.divide(iv->getValue()))
            throw IntError("Divide by zero");
        return makeStringValue(a.toCargoSpecString());
    } else if (IntStringValue* sv = dynamic_cast<IntStringValue*>(arg)) {
        // Divide cargospec by cargospec
        GCargoSpec b(sv->getValue());
        int32_t result;
        if (!a.divide(b, result))
            throw IntError("Divide by zero");
        return makeIntValue(result);
    } else {
        throw IntError::typeError(IntError::tyBaseType);
    }
}

/** Common back-end for CExtract and CRemove.
    \param cs [in/out] CargoSpec. On output, the result of CRemove(cs, s).
    \param total [out] On output, result of CExtract(cs, s).
    \param s [in] Type specification */
static void
doCExtractRemove(GCargoSpec& cs, int32_t& total, const string_t& s)
{
    total = 0;
    for (string_t::size_type i = 0; i < s.size(); ++i) {
        if (uint8_t(s[i]) > ' ') {
            int ty = GCargoSpec::charToCargoType(s[i]);
            if (ty < 0)
                throw IntError("Invalid cargo type");
            total += cs.get(GCargoType(ty));
            cs.set(GCargoType(ty), 0);
        }
    }
}

/** Implementation of the "CExtract" function. */
IntValue*
IFCExtract(IntArgBlock& args)
{
    args.checkArgumentCount(2);

    // Get first arg
    GCargoSpec cs;
    if (!checkCargoSpecArg(cs, args.getNext()))
        return 0;

    // Get second arg
    string_t s;
    if (!checkStringArg(s, args.getNext()))
        return 0;

    // Compute result
    int32_t result;
    doCExtractRemove(cs, result, s);

    return makeIntValue(result);
}

/** Implementation of the "CMul" function. */
IntValue*
IFCMul(IntArgBlock& args)
{
    args.checkArgumentCount(2);

    // Get first arg
    GCargoSpec cs;
    if (!checkCargoSpecArg(cs, args.getNext()))
        return 0;

    // Get second arg
    int32_t n;
    if (!checkIntArg(n, args.getNext()))
        return 0;

    cs *= n;
    return makeStringValue(cs.toCargoSpecString());
}

/** Implementation of the "CRemove" function. */
IntValue*
IFCRemove(IntArgBlock& args)
{
    args.checkArgumentCount(2);

    // Get first arg
    GCargoSpec cs;
    if (!checkCargoSpecArg(cs, args.getNext()))
        return 0;

    // Get second arg
    string_t s;
    if (!checkStringArg(s, args.getNext()))
        return 0;

    // Compute result
    int32_t tmp;
    doCExtractRemove(cs, tmp, s);

    return makeStringValue(cs.toCargoSpecString());
}

/** Implementation of the "CSub" function. */
IntValue*
IFCSub(IntArgBlock& args)
{
    args.checkArgumentCountAtLeast(2);

    GCargoSpec dif;
    if (!checkCargoSpecArg(dif, args.getNext()))
        return 0;

    while (args.getNumArgs() > 0) {
        GCargoSpec subtr;
        if (!checkCargoSpecArg(subtr, args.getNext()))
            return 0;
        dif -= subtr;
    }

    dif.doSupplySale();

    return makeStringValue(dif.toCargoSpecString());
}

/** Implementation of the "AddCommand" command. */
void
IFAddCommand(IntExecutionContext& /*exc*/, IntArgBlock& args)
{
    args.checkArgumentCount(1);

    // Fetch argument
    string_t cmdStr;
    if (!checkStringArg(cmdStr, args.getNext()))
        return;

    // Do it
    if (!haveDisplayedTurn() || !getPlayerId())
        throw IntError(GError::eUser);

    // Parse command
    GCommand* cmd = GCommand::parseCommand(cmdStr, true);
    if (!cmd)
        throw IntError("Invalid command");

    getDisplayedTurn().getCommands(getPlayerId()).setCommand(cmd);
}

void
IFNewCircle(IntExecutionContext& /*exc*/, IntArgBlock& args)
{
    // NewCircle x,y,radius,[color,tag,expire]
    args.checkArgumentCount(3, 6);

    int32_t x, y, radius;
    int32_t color = 9;
    int32_t tag = 0;
    int32_t expire = -1;
    if (!checkIntArg(x, args.getNext(), 0, 10000)
        || !checkIntArg(y, args.getNext(), 0, 10000)
        || !checkIntArg(radius, args.getNext(), 0, 5000))
        return;
    checkIntArg(color, args.getNext(), 0, GDrawing::NUM_USER_COLORS);
    checkIntArg(tag, args.getNext(), 0, 0xFFFF);
    checkIntArg(expire, args.getNext(), -1, 0x7FFF);

    // Do it
    GDrawing drawing(GPoint(x, y), CircleDrawing);
    drawing.setCircleRadius(radius);
    drawing.setColor(GDrawing::getColorFromUserColor(color));
    drawing.setTag(tag);
    drawing.setExpire(expire);

    getDisplayedTurn().getCurrentUniverse().getDrawings().add(drawing);
    getDisplayedTurn().getCurrentUniverse().markChanged();
    getDisplayedTurn().getChartBBox().addDrawing(drawing);
}

static void
doLineRectangle(IntArgBlock& args, GDrawingType drawingType, bool normalizeCoords)
{
    // <command> x1,y1,x2,y2[,color,tag,expire]
    args.checkArgumentCount(4, 7);

    int32_t x1, y1, x2, y2;
    int32_t color = 9;
    int32_t tag = 0;
    int32_t expire = -1;

    if (!checkIntArg(x1, args.getNext(), 0, 10000)
        || !checkIntArg(y1, args.getNext(), 0, 10000)
        || !checkIntArg(x2, args.getNext(), 0, 10000)
        || !checkIntArg(y2, args.getNext(), 0, 10000))
        return;
    checkIntArg(color, args.getNext(), 0, GDrawing::NUM_USER_COLORS);
    checkIntArg(tag, args.getNext(), 0, 0xFFFF);
    checkIntArg(expire, args.getNext(), -1, 0x7FFF);

    // Refuse making drawing of a certain size
    if (std::abs(x2-x1) > 5000 || std::abs(y2-y1) > 5000)
        throw IntError::rangeError();

    // Normalize coordinates if needed
    GPoint a(x1, y1);
    GPoint b(x2, y2);
    if (normalizeCoords)
        b = b.getNearestTo(a);

    // Draw it
    GDrawing drawing(a, drawingType);
    drawing.setPos2(b);
    drawing.setColor(GDrawing::getColorFromUserColor(color));
    drawing.setTag(tag);
    drawing.setExpire(expire);

    getDisplayedTurn().getCurrentUniverse().getDrawings().add(drawing);
    getDisplayedTurn().getCurrentUniverse().markChanged();
    getDisplayedTurn().getChartBBox().addDrawing(drawing);
}

void
IFNewRectangle(IntExecutionContext& /*exc*/, IntArgBlock& args)
{
    doLineRectangle(args, RectangleDrawing, true);
}

void
IFNewRectangleRaw(IntExecutionContext& /*exc*/, IntArgBlock& args)
{
    doLineRectangle(args, RectangleDrawing, false);
}

void
IFNewLine(IntExecutionContext& /*exc*/, IntArgBlock& args)
{
    doLineRectangle(args, LineDrawing, true);
}

void
IFNewLineRaw(IntExecutionContext& /*exc*/, IntArgBlock& args)
{
    doLineRectangle(args, LineDrawing, false);
}

void
IFNewMarker(IntExecutionContext& /*exc*/, IntArgBlock& args)
{
    // NewMarker x,y,typ[,color,text,tag,expire]
    args.checkArgumentCount(3, 7);

    int32_t x, y, type;
    int32_t color = 9;
    int32_t tag = 0;
    int32_t expire = -1;
    string_t text;
    if (!checkIntArg(x, args.getNext(), 0, 10000)
        || !checkIntArg(y, args.getNext(), 0, 10000)
        || !checkIntArg(type, args.getNext(), 0, NUM_USER_MARKERS-1))
        return;
    checkIntArg(color, args.getNext(), 0, GDrawing::NUM_USER_COLORS);
    checkStringArg(text, args.getNext());
    checkIntArg(tag, args.getNext(), 0, 0xFFFF);
    checkIntArg(expire, args.getNext(), -1, 0x7FFF);

    // Draw it
    GDrawing drawing(GPoint(x, y), MarkerDrawing);
    drawing.setMarkerKind(type);
    drawing.setColor(GDrawing::getColorFromUserColor(color));
    drawing.setTag(tag);
    drawing.setExpire(expire);
    drawing.setComment(text);

    getDisplayedTurn().getCurrentUniverse().getDrawings().add(drawing);
    getDisplayedTurn().getCurrentUniverse().markChanged();
    getDisplayedTurn().getChartBBox().addDrawing(drawing);
}

void
IFSaveGame(IntExecutionContext& /*exc*/, IntArgBlock& args)
{
    args.checkArgumentCount(0);
    if (GGameDataStorage::instance)
        GGameDataStorage::instance->save(getCurrentTurn());
    else
        throw IntError("No game loaded");
}

namespace {
    class SelectionLoadAskUI : public AnswerProvider {
     public:
        bool ask(int, string_t q);
    };
}

bool
SelectionLoadAskUI::ask(int, string_t q)
{
    /* This is wild. We have to ask the user a question. To do that,
       we need to call UI.Message, because only that knows whether
       we're in the GUI or on the console. As an added bonus, it is
       implemented as a script on the console, so we cannot just look
       up a symbol and call that natively. So, we have to create a
       fully-fledged (temporary) process running a newly-made BCO.

          pushlit question
          pushlit title
          pushlit "Yes No"
          pushvar "UI.MESSAGE"
          procind 3
          pushloc "UI.RESULT"
    */
    IntStringValue svq((format(_("%s. Do you want to load this file?"), q)));
    IntStringValue svt(_("Load Selection"));
    IntStringValue svb(_("Yes No"));
    IntBCORef bco = new IntBytecodeObject();
    bco->setIsProcedure(false);
    bco->addLocalVariable("UI.RESULT");
    bco->addPushLiteral(&svq);
    bco->addPushLiteral(&svt);
    bco->addPushLiteral(&svb);
    bco->addVariableReferenceInstruction(IntOpcode::maPush, "UI.MESSAGE", IntCompilationContext());
    bco->addInstruction(IntOpcode::maIndirect, IntOpcode::miIMCall, 3);
    bco->addVariableReferenceInstruction(IntOpcode::maPush, "UI.RESULT", IntCompilationContext());

    IntExecutionContext exc("Temporary: SelectionLoad");
    IntExecutionFrame* frame = new IntExecutionFrame(bco);
    frame->want_result = true;
    exc.pushNewFrame(frame);

    return runTemporaryProcess(exc) && getBoolValue(exc.getResult()) > 0;
}

void
IFSelectionLoad(IntExecutionContext& /*exc*/, IntArgBlock& args)
{
    // SelectionLoad "f[TAMU#]"
    //   timeless, all, merge, UI
    TextFile* tf;
    int32_t flags = 0;
    int32_t layer = GMultiSelection::getCurrentSelectionLayer();

    args.checkArgumentCount(1, 2);
    if (!checkFileArg(tf, args.getNext()))
        return;
    checkFlagArg(flags, &layer, args.getNext(), "TAMU");
    if (layer < 0 || layer >= GMultiSelection::NUM_SELECTION_LAYERS)
        throw IntError::rangeError();

    int translatedFlags = 0;
    if (flags & 4) {
        /* M = Merge */
        translatedFlags |= selio_Merge;
    }
    if (flags & 8) {
        /* U = User interface */
        SelectionLoadAskUI a;
        loadSelection(*tf, layer, translatedFlags, a);
    } else {
        /* Not user interface, so evaluate other two flags as well */
        if (flags & 1) {
            /* T = Timeless */
            translatedFlags |= selio_Timeless;
        }
        if (flags & 2) {
            /* A = All */
            translatedFlags |= selio_AcceptAll;
        }
        translatedFlags |= selio_AcceptCurrent;
        translatedFlags |= selio_AcceptSingle;

        class Answer : public AnswerProvider {
         public:
            bool ask(int qid, string_t /*q*/) {
                if (qid & selio_Timeless) {
                    throw IntError("Stale file");
                }
                throw IntError("File doesn't match requested content");
            }
        };
        Answer a;
        loadSelection(*tf, layer, translatedFlags, a);
    }
}

void
IFSelectionSave(IntExecutionContext& /*exc*/, IntArgBlock& args)
{
    // SelectionSave "f[T#]"
    TextFile* tf;
    int32_t flags = 0;
    int32_t layer = -1;

    args.checkArgumentCount(1, 2);
    if (!checkFileArg(tf, args.getNext()))
        return;
    checkFlagArg(flags, &layer, args.getNext(), "T");
    if (layer < -1 || layer >= GMultiSelection::NUM_SELECTION_LAYERS)
        throw IntError::rangeError();

    saveSelection(*tf, layer, flags ? selio_Timeless : 0);
}

void
IFCCSelectionExec(IntExecutionContext& /*exc*/, IntArgBlock& args)
{
    args.checkArgumentCount(2);

    int32_t layer;
    string_t code;
    if (!checkIntArg(layer, args.getNext(), 0, GMultiSelection::NUM_SELECTION_LAYERS)
        || !checkStringArg(code, args.getNext()))
        return;

    if (layer == 0)
        layer = GMultiSelection::getCurrentSelectionLayer();
    else
        --layer;

    GMultiSelection::executeSelectionExpression(layer, code);
}
