/**
  *  \file tools/c2mgrep.cc
  *  \brief "Message Grep" utility
  *
  *  (c) 2001-2011 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <cstdlib>
#include <vector>
#include <cstring>
#include "arch/platform.h"
#include "cpluslib/format.h"
#include "cpluslib/string.h"
#include "game/checksum.h"
#include "game/msg.h"
#include "game/racename.h"
#include "game/rst.h"
#include "game/struct.h"
#include "io/io.h"
#include "io/streams.h"
#include "util/console.h"
#include "util/except.h"
#include "util/modinit.h"
#include "util/nls.h"
#include "util/recode.h"
#include "version.h"
#include "game/inbox.h"

#define VERSION PCC2_VERSION

static const char* progname;

/** Exit with error. */
static void
die(string_t str)
{
    console.write(LOG_ERROR, format("%s: %s", progname, str));
    std::exit(1);
}

/** Exit with help message. */
static void
help()
{
    console << format(_("PCC2 Message Search v%s - (c) 2011 Stefan Reuther\n"), VERSION)
            << format(_("\n"
                        "Usage:\n"
                        "  %s [-h]\n"
                        "  %$0s [-opts] \"search string\" [-type] files...\n\n"
                        "Options:\n"
                        "  -c           Case-sensitive\n"
                        "  -C CHARSET   Select character set\n"
                        "\n"
                        "Type options apply to all subsequent file names:\n"
                        "  -r           Result files\n"
                        "  -t           Turn files\n"
                        "  -m           Message inbox (MDATA)\n"
                        "  -d           Dosplan outbox (MESS)\n"
                        "  -w           Winplan outbox (MESS35)\n"
                        "  -a           VPA database\n"
                        "  -A           Auto-detect (default)\n"
                        "\n"
                        "Report bugs to <Streu@gmx.de>\n"),
                      progname);
    console.flush();
    std::exit(0);
}

/************************** File Identification **************************/

enum FileType {
    UnknownFile,
    TurnFile,
    ResultFile,
    OutboxFile,
    Outbox35File,
    InboxFile,
    VpaDatabaseFile
};

static bool
isTurn(const char* data, size_t size, long total)
{
    if (size <= TTurnHeader::size)
        return false;

    /* Validate ranges. 100000 is an estimate how many commands a turn file can
       possibly contain. MGREP 1.3d uses 5000, which is consistent with Host, but
       a turn can legally contain more commands. */
    /* FIXME: this does not recognize Taccom. Should it? */
    TTurnHeader hdr;
    getStructure(data, hdr);
    if (hdr.player <= 0 || hdr.player > 11)
        return false;
    if (hdr.num_commands < 0 || hdr.num_commands > 100000)
        return false;
    if (getChecksum(hdr.timestamp, sizeof(hdr.timestamp)) != hdr.time_checksum)
        return false;
    if (total < TTurnHeader::size || total - TTurnHeader::size < hdr.num_commands*4)
        return false;
    return true;
}

static bool
isResult(const char* data, size_t size, long total)
{
    if (size <= 8*4)
        return false;

    /* Check pointers in standard range */
    for (int i = 0; i < 8; ++i) {
        int32_t v = getInt32(data + 4*i);
        if (v < 20 || v > total)
            return false;
    }
    return true;
}

static bool
isOutboxFile(const char* data, size_t size, long total)
{
    if (size < 2)
        return false;

    /* Check count, and that the header fits in the file completely. */
    uint16_t nmess = getUint16(data);
    long minsize = nmess*long(TOutgoingMessageHeader::size) + 2;
    if (minsize > total)
        return false;

    /* Figure out how many headers fit in our preview */
    size_t max = (size - 2) / TOutgoingMessageHeader::size;
    if (nmess > max)
        nmess = max;

    /* Check all headers. Length and position must be reasonable,
       sender numbers must agree, receiver must be in range. */
    int sender = 0;
    for (uint16_t i = 0; i < nmess; ++i) {
        TOutgoingMessageHeader omh;
        getStructure(data + 2 + TOutgoingMessageHeader::size*i, omh);
        if (omh.address < minsize || omh.address > total)
            return false;
        if (omh.length < 0 || omh.length > 10000)
            return false;
        if (omh.to <= 0 || omh.to > 12)
            return false;
        if (sender == 0) {
            if (omh.from <= 0 || omh.from > 11)
                return false;
            sender = omh.from;
        } else {
            if (omh.from != sender)
                return false;
        }
    }
    return true;
}

static bool
isOutbox35File(const char* data, size_t size, long total)
{
    /* File consists of a word with a count, 18 garbage bytes, and 13
       flag bytes, where the flag bytes are actually part of a
       repeating structure. Check that count and flag bytes are valid. 
       It makes no sense to check the repeating structure, as our
       preview will probably be too small anyway. */
    int16_t count = getInt16(data);
    if (size < 2+18+13)
        return false;
    if (count < 0 || count > 10000)
        return false;
    if (total < 100)
        return false;
    for (int i = 0; i < 13; ++i)
        if (data[20+i] != '0' && data[20+i] != '1')
            return false;
    return true;
}

static bool
isInbox(const char* data, size_t size, long total)
{
    if (size < 2)
        return false;

    /* Check count, and that the header fits in the file completely. */
    uint16_t nmess = getUint16(data);
    long minsize = nmess*long(TIncomingMessageHeader::size) + 2;
    if (minsize > total)
        return false;

    /* Figure out how many headers fit in our preview */
    size_t max = (size - 2) / TIncomingMessageHeader::size;
    if (nmess > max)
        nmess = max;

    /* Check all headers. Length and position must be reasonable. */
    for (uint16_t i = 0; i < nmess; ++i) {
        TIncomingMessageHeader imh;
        getStructure(data + 2 + TIncomingMessageHeader::size*i, imh);
        if (imh.address < minsize || imh.address > total)
            return false;
        if (imh.length < 0 || imh.length > 10000)
            return false;
    }
    return true;
}

static bool
isVpaDatabase(const char* data, size_t size, long total)
{
    static const char dbsig[] = "VPA Database\015\012\006";
    static const size_t dbsiglen = sizeof(dbsig)-1;

    if (total < 128 || size < dbsiglen)
        return false;
    return (std::memcmp(data, dbsig, dbsiglen) == 0);
}

static FileType
identifyFile(Stream& s)
{
    /* Read first bytes of the file */
    char tmp[512];
    size_t n   = s.read(tmp, sizeof(tmp));
    long total = s.getSize();
    s.seek(0);

    /* Apply all checkers, most reliable first */
    if (isVpaDatabase(tmp, n, total))
        return VpaDatabaseFile;
    if (isTurn(tmp, n, total))
        return TurnFile;
    if (isResult(tmp, n, total))
        return ResultFile;
    if (isOutboxFile(tmp, n, total))
        return OutboxFile;
    if (isOutbox35File(tmp, n, total))
        return Outbox35File;
    if (isInbox(tmp, n, total))
        return InboxFile;
    return UnknownFile;
}

/********************************* Search ********************************/

static void
searchInbox(Stream& s, const string_t& searchString, bool optCaseSense)
{
    /* This is simple, but inefficient */
    GInbox inbox;
    inbox.loadInbox(s, 0);
    for (int i = 0, e = inbox.getCount(); i < e; ++i) {
        string_t t = inbox.getText(i);
        if (!optCaseSense)
            t = strUCase(t);
        if (t.find(searchString) != t.npos) {
            console << format("=== %s, message %d ===\n", s.getName(), i+1)
                    << inbox.getText(i)
                    << "\n";
        }
    }
}

static void
searchResult(Stream& s, const string_t& searchString, bool optCaseSense)
{
    GResultFile rst(s);
    s.seek(rst.getSectionOffset(GResultFile::sec_Messages));
    searchInbox(s, searchString, optCaseSense);
}

static void
searchFile(const char* fname, string_t searchString, bool optCaseSense, FileType optFileType)
{
    try {
        Ptr<Stream> file = Stream::openFile(fname, Stream::C_READ);
        if (!optCaseSense)
            searchString = strUCase(searchString);
        if (optFileType == UnknownFile)
            optFileType = identifyFile(*file);
        switch (optFileType) {
         case TurnFile:
            // FIXME:
            break;
         case ResultFile:
            searchResult(*file, searchString, optCaseSense);
            break;
         case OutboxFile:
         case Outbox35File:
            // FIXME
            break;
         case InboxFile:
            searchInbox(*file, searchString, optCaseSense);
            break;
         case VpaDatabaseFile:
         case UnknownFile:
            console.write(LOG_ERROR, format("%s: unknown file format", fname));
            break;
        }
    }
    catch (FileProblemException& ex) {
        console.write(LOG_ERROR, format("%s: %s", ex.getFileName(), ex.what()));
    }
    catch (std::exception& ex) {
        console.write(LOG_ERROR, format("%s: %s", fname, ex.what()));
    }
}

static const ModuleInitializer*const mgrep_modules[] = {
    mod_base_console,
    mod_racenames,
    0
};

int
main(int /*argc*/, char** argv)
{
    /* Arguments */
    const char* searchString = 0;
    bool hadFiles = false;
    bool optCaseSense = false;
    FileType optFileType = UnknownFile;

    /* Initialize */
    performInitialisation(OnStartup, mgrep_modules);
    argv = Platform::convertArgumentsToUtf8(argv);
    progname = argv[0];

    /* Parse and search */
    while (const char* arg = *++argv) {
        if (*arg == '-') {
            while (char c = *++arg) {
                switch (c) {
                 case 'h':
                    help();
                    break;
                 case 'C':
                    if (! *++argv)
                        die(_("option `-C' needs an argument (the character set)"));
                    {
                        CharacterSet cs(CharacterSet::getCharsetByName(*argv));
                        if (!cs.isValid())
                            die(_("the specified character set is not known"));
                        setGameCharacterSet(cs);
                    }
                 case 'r':
                    optFileType = ResultFile;
                    break;
                 case 't':
                    optFileType = TurnFile;
                    break;
                 case 'm':
                    optFileType = InboxFile;
                    break;
                 case 'd':
                    optFileType = OutboxFile;
                    break;
                 case 'w':
                    optFileType = Outbox35File;
                    break;
                 case 'a':
                    optFileType = VpaDatabaseFile;
                    break;
                 case 'A':
                    optFileType = UnknownFile;
                    break;
                 case 'c':
                    optCaseSense = true;
                    break;
                 default:
                    die(format(_("invalid option specified. Use `%s -h' for help"), progname));
                    break;
                }
            }
        } else if (!searchString) {
            searchString = arg;
        } else {
            searchFile(arg, searchString, optCaseSense, optFileType);
            hadFiles = true;
        }
    }

    if (!searchString) {
        die(format(_("no search string specified. Use `%s -h' for help"), progname));
    }
    if (!hadFiles) {
        die(format(_("no file name specified. Use `%s -h' for help"), progname));
    }
    return 0;
}
